#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/buffer_head.h>
#include <linux/string.h>
#include <linux/errno.h>
#include <linux/cellfs_fs.h>
#include "cellfs.h"
#include "mdt.h"

/*----------------------------------------------------------------------------*/
/* Structs of blocks in atime file                                            */
/*----------------------------------------------------------------------------*/

#pragma pack(1)
struct cellfs_atime_block_header {
    __u64 count;
};

/* XXX is block size 512 or 4096? */
#define CELLFS_ENTRIES_IN_ATIME_BLOCK ((512 - sizeof(struct cellfs_atime_block_header)) / sizeof(struct timespec))

struct cellfs_atime_block {
    struct cellfs_atime_block_header header;
    struct timespec entries[CELLFS_ENTRIES_IN_ATIME_BLOCK];
};
#pragma pack()

/*----------------------------------------------------------------------------*/
/* Functions for getting a mapped block to atime file                         */
/*----------------------------------------------------------------------------*/

static void cellfs_atime_block_init(struct inode *atime,
                                   struct buffer_head *bh,
                                   void *kaddr)
{
	struct cellfs_atime_block * atime_block = kaddr + bh_offset(bh);
    memset(atime_block, 0, sizeof(*atime_block));
}

static inline unsigned long cellfs_atime_get_block_number(unsigned long inode_num)
{
    return inode_num / CELLFS_ENTRIES_IN_ATIME_BLOCK;
}

static inline struct timespec * cellfs_atime_get_entry_in_block(struct cellfs_atime_block * atime_block,
                                                              unsigned long inode_num)
{
    return &atime_block->entries[inode_num % CELLFS_ENTRIES_IN_ATIME_BLOCK];
}

/**
 * cellfs_atime_get_block - get a block
 * @atime: inode of atime file
 * @inode_num: inode number
 * @block_out: pointer to the block
 * @block_bh_out: pointer to a buffer head
 *
 * Description: cellfs_atime_get_block() acquires the block hosting the atime of
 * @inode_num. A new block is created if doesn't exist.
 *
 * Return Value: On success, 0 is returned, and the block and the
 * buffer head of the block on which the block is located are stored in
 * the place pointed by @block_out and @block_bh_out, respectively. On error, one of the
 * following negative error codes is returned.
 *
 * %-EIO - I/O error.
 *
 * %-ENOMEM - Insufficient amount of memory available.
 */
static int cellfs_atime_get_block(struct inode * atfile,
                                 __u64 inode_num,
                                 struct cellfs_atime_block ** block_out,
                                 struct buffer_head ** block_bh_out)
{
	struct buffer_head *block_bh;
	struct cellfs_atime_block * atime_block;
	void *kaddr;
	int ret;

	down_write(&CELLFS_MDT(atfile)->mi_sem);

    /* Get the block from MDT */
    ret = cellfs_mdt_get_block(atfile,
                              cellfs_atime_get_block_number(inode_num),
                              1, cellfs_atime_block_init, &block_bh);
    if (ret < 0) {
		goto out;
    }

    /* Map the block to lower memory */
    kaddr = kmap(block_bh->b_page);
    atime_block = (struct cellfs_atime_block *)(kaddr + bh_offset(block_bh));

    /* Make sure the block will be flushed when we finish */
    mark_buffer_dirty(block_bh);
    cellfs_mdt_mark_dirty(atfile);

    /* Return the atime_block and buffer head to the user */
    if (block_out != NULL) {
		*block_out = atime_block;
    }
    *block_bh_out = block_bh;

 out:
	up_write(&CELLFS_MDT(atfile)->mi_sem);
	return ret;
}

/**
 * cellfs_atime_put_block - put an atime block
 * @bh: buffer head returned by a successful call to cellfs_atime_get_block
 */
static void cellfs_atime_put_block(struct buffer_head * block_bh)
{
	kunmap(block_bh->b_page);
	brelse(block_bh);
}

/*----------------------------------------------------------------------------*/
/* External interfaces for setting/getting/deleting atime                     */
/*----------------------------------------------------------------------------*/

int cellfs_atime_fill_inode(struct inode * atfile, struct inode * inode)
{
    int err;
    struct cellfs_atime_block * atime_block;
    struct buffer_head * block_bh;
    struct timespec * atime;

    if ((NULL == atfile) || (atfile == inode)) {
        inode->i_atime = inode->i_mtime;
        return 0;
    }

    err = cellfs_atime_get_block(atfile, inode->i_ino, &atime_block, &block_bh);
    if (err < 0) {
        return err;
    }

    atime = cellfs_atime_get_entry_in_block(atime_block, inode->i_ino);
    if ((0 == atime->tv_sec) && (0 == atime->tv_nsec)) {
        /* time is invalid, set to valid time and increase count */
        *atime = inode->i_mtime;
        ++atime_block->header.count;
    }

    inode->i_atime = *atime;
    cellfs_atime_put_block(block_bh);
    return 0;
}

int cellfs_atime_update_from_inode(struct inode * atfile, struct inode * inode)
{
    int err;
    struct cellfs_atime_block * atime_block;
    struct buffer_head * block_bh;
    struct timespec * atime;

    err = cellfs_atime_get_block(atfile, inode->i_ino, &atime_block, &block_bh);
    if (err < 0) {
        return err;
    }

    atime = cellfs_atime_get_entry_in_block(atime_block, inode->i_ino);
    *atime = inode->i_atime;

    cellfs_atime_put_block(block_bh);
    return 0;
}

int cellfs_atime_delete_inode_entry(struct inode * atfile, struct inode * inode)
{
    int err;
    int needs_release;
    struct cellfs_atime_block * atime_block;
    struct buffer_head * block_bh;
    struct timespec * atime;

    err = cellfs_atime_get_block(atfile, inode->i_ino, &atime_block, &block_bh);
    if (err < 0) {
        return err;
    }

    atime = cellfs_atime_get_entry_in_block(atime_block, inode->i_ino);
    atime->tv_sec = 0;
    atime->tv_nsec = 0;
    --atime_block->header.count;
    needs_release = (0 == atime_block->header.count);

    cellfs_atime_put_block(block_bh);

    if (needs_release) {
        cellfs_mdt_delete_block(atfile, cellfs_atime_get_block_number(inode->i_ino));
    }

    return 0;
}

int cellfs_atime_read(struct super_block *sb,
                     struct cellfs_inode *raw_inode,
                     struct inode **inodep)
{
	struct inode *atfile;
	int err;

    atfile = cellfs_iget_locked(sb, NULL, CELLFS_ATIME_INO);
	if (unlikely(!atfile))
		return -ENOMEM;
	if (!(atfile->i_state & I_NEW))
		goto out;
	
	err = cellfs_mdt_init(atfile, CELLFS_MDT_GFP, 0);
	if (err)
		goto failed;

	cellfs_mdt_set_entry_size(atfile,
                             sizeof(struct cellfs_atime_block),
                             sizeof(struct cellfs_atime_block_header));

	err = cellfs_read_inode_common(atfile, raw_inode);
	if (err)
		goto failed;

 out:
    *inodep = atfile;
	return 0;
 failed:
    iget_failed(atfile);
	return err;
}